import inspect
from typing import Any, Callable, Dict, List

from pydantic import create_model

from actionai.exceptions import ActionAIException


def create_json_schema_for_function_input(
    fn: Callable, skip_keys: List[str]
) -> Dict[str, Any]:
    """Create json schema for a function's input

    Args:
        fn (Callable): The function
        skip_keys (List[str]): The arguments to skip (context keys)

    Returns:
        Dict[str, Any]: The json schema
    """
    signature = inspect.signature(fn)
    parameters = signature.parameters

    model_fields = {}

    for param in parameters.values():
        if param.name in skip_keys:
            continue

        default_param = ...
        if param.default is not inspect._empty:
            default_param = param.default

        annotation = param.annotation
        if annotation is param.empty and default_param is not ...:
            annotation = type(default_param)
        if annotation is param.empty:
            raise ActionAIException(
                f"parameter '{param.name}' has missing type annotation"
            )

        model_fields[param.name] = (annotation, default_param)

    DynModel = create_model(fn.__name__, **model_fields)  # type: ignore
    return DynModel.model_json_schema()



import json
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, cast

import openai

from actionai.exceptions import ActionAIException
from actionai.json_schema import create_json_schema_for_function_input
from actionai.types import (
    OPENAI_MODELS,
    ChatResponse,
    ChatResponseMessage,
    Message,
    OpenAIFunction,
)


@dataclass
class ActionAIFunction:
    fn: Callable
    name: str
    description: str
    input_schema: Any

    def to_openai_function(self) -> OpenAIFunction:
        return {
            "name": self.name,
            "description": self.description,
            "parameters": self.input_schema,
        }


class ActionAI:
    DEFAULT_SYSTEM_PROMPT = "Don't make assumptions about what values to \
                    plug into functions. Ask for clarification if a user \
                        request is ambiguous."

    def __init__(
        self,
        openai_api_key: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        model: OPENAI_MODELS = "gpt-3.5-turbo-0613",
        system_prompt: Optional[str] = None,
        temperature: int = 1,
    ) -> None:
        """
        Args:
            openai_api_key (Optional[str]): If not set, defaults \
                to `OPENAI_API_KEY` environment variable.

            context (Optional[Dict[str, Any]]): These keys will be skipped \
                when creating json schema for the function's input. The values \
                    will be directly passed during function call.

            model (optional): The chat completion model to use. Defaults to \
                "gpt-3.5-turbo-0613".
        """
        assert (
            temperature >= 0 and temperature <= 2
        ), "Temperature must be between 0 and 2"

        if openai_api_key is not None:
            openai.api_key = openai_api_key

        self.messages: List[Message] = []
        self.context = context or {}
        self.model = model
        self.system_prompt = system_prompt or self.DEFAULT_SYSTEM_PROMPT
        self.temperature = temperature

        # Do not update these attributes directly
        self._functions: Dict[str, ActionAIFunction] = {}
        self._openai_functions: List[OpenAIFunction] = []

    def _add_system_prompt(self):
        self.messages.append(
            {
                "role": "system",
                "content": self.system_prompt,
            }
        )

    def set_messages(self, messages: List[Message]):
        self._add_system_prompt()
        self.messages = messages

    def register(self, fn: Callable):
        if fn.__name__ in self._functions:
            raise ActionAIException("function with the same name already registered")

        if fn.__doc__ is None:
            raise ActionAIException("function must have a docstring")

        action_function = ActionAIFunction(
            fn=fn,
            name=fn.__name__,
            description=fn.__doc__,
            input_schema=create_json_schema_for_function_input(
                fn=fn, skip_keys=list(self.context.keys())
            ),
        )
        self._functions[fn.__name__] = action_function
        self._openai_functions.append(action_function.to_openai_function())

    def _create_chat_completion(self) -> ChatResponse:
        response = openai.ChatCompletion.create(
            model=self.model,
            messages=self.messages,
            functions=self._openai_functions,
            function_call="auto",
            temperature=self.temperature,
        )
        response = cast(ChatResponse, response)
        response_message = response["choices"][0]["message"]
        self.messages.append(response_message)  # type: ignore

        if response_message.get("function_call") is None:
            return response

        function_response = self._execute_function(response_message)
        self.messages.append(
            {
                "role": "function",
                "name": response_message["function_call"]["name"],
                "content": json.dumps(function_response, default=str),
            }
        )

        response = openai.ChatCompletion.create(
            model=self.model, messages=self.messages, temperature=self.temperature
        )

        response = cast(ChatResponse, response)
        self.messages.append(response["choices"][0]["message"])  # type: ignore
        return response

    def _execute_function(self, response_message: ChatResponseMessage) -> Any:
        function_name = response_message["function_call"]["name"]
        fuction_to_call = self._functions[function_name].fn
        function_args = json.loads(response_message["function_call"]["arguments"])
        return fuction_to_call(**function_args, **self.context)

    def prompt(self, query: str) -> ChatResponse:
        if len(self.messages) == 0:
            self._add_system_prompt()

        self.messages.append({"role": "user", "content": query})
        return self._create_chat_completion()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['actionai']

package_data = \
{'': ['*']}

install_requires = \
['openai>=0.27.8,<0.28.0', 'pydantic>=2.0.2,<3.0.0']

setup_kwargs = {
    'name': 'actionai',
    'version': '0.0.7',
    'description': 'A small library to call local functions using openai function calling',
    'long_description': '# ActionAI\n\nA small library to run local functions using the openai function calling\n\n> **Warning**\n> This library is still in its early stages, so please use it cautiously. If you find any bugs, please create a new issue.\n\n## Install\n\n```shell\npip install actionai\n```\n\n## Usage\n\n> **Note**\n> A function must be fully typed and must have a docstring\n\n```python\n# define a new function\ndef get_current_weather(location: str, unit: str = "fahrenheit"):\n    """Function to get current weather for the given location"""\n    weather_info = {\n        "location": location,\n        "temperature": "72",\n        "unit": unit,\n        "forecast": ["sunny", "windy"],\n    }\n    return weather_info\n\n\nimport actionai\n\naction = actionai.ActionAI()\naction.register(get_current_weather)\n\nresponse = action.prompt("What is the current weather in the north pole?")\n\nprint(response["choices"][0]["message"]["content"])\n# The current weather at the North Pole is 72°F. It is sunny and windy.\n```\n\nThe openai api key will be read automatically from the `OPENAI_API_KEY` environment variable. You can pass it manually as,\n\n```python\nimport actionai\n\naction = actionai.ActionAI(openai_api_key="YOUR_KEY")\n```\n\n### Adding context\n\nSometimes your function will have variables that need to be set by the program.\n\n```python\ndef list_todos(user: str):\n    """Function to list all todos"""\n    return todos[user]\n\naction = actionai.ActionAI(context={"user": "jason"})\n```\n\nThe context keys are skipped when creating json schema. The values are directly passed at the time of the function call.\n\n### Choosing a model\n\nBy default, the completion run on the `gpt-3.5-turbo-0613` model. You can change the model using the `model` argument.\n\n```python\nimport actionai\n\naction = actionai.ActionAI(model="gpt-4")\n```\n\nYou can see the complete list of supported chat completion models [here](https://platform.openai.com/docs/models/model-endpoint-compatibility)\n\n## Demo\n\nRunning [todo example](https://github.com/amalshaji/actionai/blob/main/examples/todo.py) 👇🏻\n\n![todo demo](https://raw.githubusercontent.com/amalshaji/actionai/main/examples/demo.svg)\n\nFor more examples, check out the [examples](https://github.com/amalshaji/actionai/tree/main/examples) directory.\n',
    'author': 'Amal Shaji',
    'author_email': 'amalshajid@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Dict, List, Literal, Optional, TypedDict


class _MessageBase(TypedDict):
    role: str
    content: str


class Message(_MessageBase, total=False):
    name: Optional[str]
    function_call: Optional[str]


class _BaseChatResponseMessage(TypedDict):
    role: str
    content: Optional[str]


class ChatResponseFunctionCall(TypedDict):
    name: str
    arguments: str


class ChatResponseMessage(_BaseChatResponseMessage):
    function_call: ChatResponseFunctionCall


class ChatResponseChoices(TypedDict):
    index: int
    message: ChatResponseMessage
    finish_reason: str


class ChatResponseUsage(TypedDict):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class ChatResponse(TypedDict):
    id: str
    object: str
    created: int
    model: str
    choices: List[ChatResponseChoices]
    usage: ChatResponseUsage


class OpenAIFunction(TypedDict):
    name: str
    description: str
    parameters: Dict[str, Any]


OPENAI_MODELS = Literal[
    "gpt-4",
    "gpt-4-0613",
    "gpt-4-32k",
    "gpt-4-32k-0613",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-16k-0613",
]
